from email.Utils import parseaddr
import os
import re
import shutil
import tempfile

from bzrlib import (
    errors as bzr_errors,
    revision as _mod_revision,
    transport,
    urlutils,
    )
from bzrlib.branch import Branch
from bzrlib.plugin import load_plugins
from bzrlib.revisionspec import RevisionSpec
from celery import shared_task
from django.conf import settings
from django.contrib.auth.models import User
from django.db.models import Max
from launchpadlib.launchpad import Launchpad

from .models import (
    Project,
    Revision,
    )


def get_or_create_branch(branch_url, project_name, fetched_branches=None,
        all_revs=False):
    basedir = settings.BRANCH_CACHE_DIR
    project_dir = os.path.join(basedir, project_name)
    if os.path.exists(project_dir):
        branch_to = Branch.open(project_dir)
        update = True
        if fetched_branches is not None:
            if project_dir in fetched_branches:
                update = False
        branch_to.lock_write()
        if update:
            try:
                branch_from = Branch.open(branch_url)
                branch_from.lock_read()
                try:
                    branch_from_revs = branch_from.revision_history()
                    branch_from_revs_count = len(branch_from_revs)
                    branch_to_revno = branch_to.revno()
                    if branch_from_revs_count >= branch_to_revno:
                        revision_id = None
                    else:
                        if branch_to_revno + 1000 >= branch_from_revs_count:
                            revision_id = None
                        else:
                            revision_id = branch_from_revs[branch_to_revno + 1000]
                    if revision_id is not None:
                        print "Getting up to revision %d" % (branch_to_revno + 1000,)
                    branch_to.pull(branch_from, True, revision_id)
                finally:
                    branch_from.unlock()
            except:
                branch_to.unlock()
                raise
    else:
        branch_from = Branch.open(branch_url)
        branch_from.lock_read()
        try:
            if not os.path.exists(basedir):
                os.makedirs(basedir)
            to_transport = transport.get_transport(project_dir)
            to_transport.mkdir('.')
            revision_id = None
            branch_from_revs = branch_from.revision_history()
            branch_from_revs_count = len(branch_from_revs)
            if not all_revs and branch_from_revs_count > 1000:
                revision_id = branch_from_revs[1000]
                print "Getting up to revision 1000"
            to_dir = branch_from.bzrdir.sprout(to_transport.base,
                                        create_tree_if_local=False,
                                        source_branch=branch_from,
                                        revision_id=revision_id)
            branch_to = to_dir.open_branch()
            branch_to.lock_write()
        finally:
            branch_from.unlock()
    if fetched_branches is not None:
        fetched_branches[project_dir] = True
    return branch_to


def get_assignee(revision, lp, email_map):
    assignee = None
    email = parseaddr(revision.get_apparent_authors()[0])[1]
    if email is not None:
        if email in email_map:
            assignee = email_map[email]
        else:
            if "." not in email:
                assignee = None
            else:
                assignee = lp.people.getByEmail(email=email)
            email_map[email] = assignee
    if assignee is None:
        return None
    return assignee.name


def get_committer(name):
    person, created = User.objects.get_or_create(username=name)
    if created:
        person.save()
    return person


def get_revision_link(branch_url, revno):
    link = branch_url
    if branch_url.startswith("lp:"):
        link = branch_url.replace("lp:", "https://bazaar.launchpad.net/+branch/")
    return link + "/revision/%d" % revno


def revision_from_bzr(model_rev, project, revno, bzr_rev, lp, email_map, branch_url):
    model_rev.commit_message = bzr_rev.message
    assignee_name = get_assignee(bzr_rev, lp, email_map)
    if assignee_name is not None:
        model_rev.assignee = assignee_name
        model_rev.committer = get_committer(assignee_name)
    model_rev.link = get_revision_link(branch_url, revno)


def store_revisions(project, revisions):
    tempdir = tempfile.mkdtemp()
    try:
        lp = Launchpad.login_anonymously('deploy_tracker', service_root='production', version='1.0',
                launchpadlib_dir=tempdir)
        email_map = {}
        # Walk the revisions oldest first
        for revno, bzr_rev, sourcedep_changes in reversed(revisions):
            try:
                parent_rev = project.revision_set.filter(parent__exact=None).get(revno=revno)
                created = False
            except Revision.DoesNotExist:
                parent_rev = Revision(project=project, revno=revno, parent=None, branch="")
                created = True
            revision_from_bzr(parent_rev, project, revno, bzr_rev, lp, email_map, project.branch)
            if created:
                if project.no_qa_of_main_branch:
                    parent_rev.status = parent_rev.STATUS_OK
                else:
                    parent_rev.status = parent_rev.STATUS_UNKNOWN
            parent_rev.save()
            if sourcedep_changes:
                for revno, bzr_rev, branch in reversed(sourcedep_changes):
                    try:
                        model_rev = project.revision_set.filter(branch__exact=branch).exclude(
                                parent__exact=None).get(revno=revno)
                        created = False
                    except Revision.DoesNotExist:
                        model_rev = Revision(project=project, revno=revno, branch=branch)
                        created = True
                    revision_from_bzr(model_rev, project, revno, bzr_rev, lp, email_map, branch)
                    if created:
                        model_rev.status = model_rev.STATUS_UNKNOWN
                    model_rev.parent = parent_rev
                    model_rev.save()
    finally:
        shutil.rmtree(tempdir)


def parse_sourcedeps_line(line):
    line = line.strip()
    if not line or line.startswith("#"):
        return None, None, None
    local_path, branch = re.split("\s+", line, 1)
    branch = re.split("\s+", branch, 1)[0]
    revision = None
    if ";revno=" in branch:
        branch, revision = branch.rsplit(";revno=", 1)
    return local_path, branch, revision


def get_sourcedeps_content(tree, path):
    file_id = tree.path2id(path)
    if file_id is None:
        print "No file at %s" % path
        return {}
    content = {}
    for line in tree.get_file_lines(file_id, path=path):
        local_path, branch, revision = parse_sourcedeps_line(line)
        if local_path is not None:
            content[local_path] = (branch, revision)
    return content


def fixup_branch_location(branch_location):
    """Adjust a branch location to remove special hosts."""
    return "lp:" + urlutils.parse_url(branch_location)[-1].lstrip("/")


def revid_and_revno_from_revspec(branch, revspec):
    if revspec is None:
        return None, None
    rev_id = RevisionSpec.from_string(revspec).as_revision_id(branch)
    revno = branch.revision_id_to_dotted_revno(rev_id)[0]
    return rev_id, revno


def sourcedeps_diff(old, new, branch_id, parent_branch_url,
        parent_branch_revision, fetched_branches=None):
    diff = []
    print "Old: %s\nNew: %s" % (old, new)
    for local_path, branch_info in new.items():
        print "Considering %s" % local_path
        old_revision = None
        max_revs = -1
        new_branch_url, new_revision = branch_info
        new_branch_url = fixup_branch_location(new_branch_url)
        print "current sourcedeps line: branch=%s revno=%s" % (new_branch_url, new_revision)
        if local_path in old:
            print "Present in old"
            old_branch_url, old_revision = old[local_path]
            old_branch_url = fixup_branch_location(old_branch_url)
            print "previous sourcedeps line: branch=%s revno=%s" % (old_branch_url, old_revision)
            if old_branch_url != new_branch_url:
                print "Changed branch from old, so doing one rev"
                max_revs = 1
            elif old_revision is None and new_revision is None:
                if parent_branch_url == new_branch_url:
                    print "Self-referring sourcedep, doing one rev"
                    max_revs = 1
                    new_revision = parent_branch_revision
                else:
                    print "No idea about revisions"
                    continue
            elif old_revision is None or new_revision is None:
                print "Revno added or removed, so doing one rev"
                max_revs = 1
        else:
            print "Not in old"
            max_revs = 1
        new_branch_id = branch_id + local_path.replace("/", "_")
        branch = get_branch(new_branch_url, new_branch_id, allow_missing=True,
                fetched_branches=fetched_branches, all_revs=True)
        if branch is None:
            print "Branch couldn't be found, skipping"
            continue
        try:
            graph = branch.repository.get_graph()
            try:
                old_rev_id, old_revno = revid_and_revno_from_revspec(branch,
                        old_revision)
            except bzr_errors.InvalidRevisionSpec:
                # Can't determine old revision, so just take one
                # revision
                max_revs = 1
            new_rev_id, new_revno = revid_and_revno_from_revspec(branch,
                    new_revision)
            if max_revs == -1:
                if old_rev_id == new_rev_id:
                    print "Same in old and new, so skipping"
                    continue
                merged_rev = graph.find_lefthand_merger(old_rev_id, new_rev_id)
                if merged_rev is None:
                    print "Went back to older revision, so doing one rev"
                    max_revs = 1
            revisions = get_revisions(branch, new_branch_id, parent_branch_url,
                    start_revno=old_revno, end_revno=new_revno, max_revs=max_revs,
                    allow_missing=True)
            diff.extend([(revno, rev, new_branch_url) for revno, rev, _ in revisions])
        finally:
            branch.unlock()
    return diff


def sourcedeps_diff_revisions(branch, current_revision, previous_revision, sourcedeps_path,
        branch_id, parent_branch_url, fetched_branches=None):
    current_sourcedeps_content = get_sourcedeps_content(
            branch.repository.revision_tree(current_revision.revision_id),
            sourcedeps_path)
    previous_sourcedeps_content = get_sourcedeps_content(
            branch.repository.revision_tree(previous_revision.revision_id),
            sourcedeps_path)
    return sourcedeps_diff(previous_sourcedeps_content,
            current_sourcedeps_content, branch_id, parent_branch_url,
            current_revision.revision_id,
            fetched_branches=fetched_branches)


def get_previous_revision(current_revision, revno, branch):
    previous_revno = None
    if not len(current_revision.parent_ids):
        previous_revision = None
    else:
        next_revision_id = current_revision.parent_ids[0]
        previous_revision = branch.repository.get_revision(next_revision_id)
        previous_revno = revno - 1
    return previous_revision, previous_revno


def get_branch(branch_url, branch_id, allow_missing=False,
        fetched_branches=None, all_revs=False):
    print "Getting branch %s" % branch_url
    try:
        branch = get_or_create_branch(branch_url, branch_id,
                fetched_branches=fetched_branches, all_revs=all_revs)
    except (bzr_errors.NotBranchError, bzr_errors.InvalidURL), e:
        if not allow_missing:
            raise
        print "Branch not found: %s" % str(e)
        return None
    print "Got branch"
    return branch


def get_revisions(branch, branch_id, parent_branch_url, start_revno=None,
        end_revno=None, sourcedeps_path=None, max_revs=-1, allow_missing=False,
        fetched_branches=None):
    if start_revno is not None and branch.revno() <= start_revno:
        return []
    if start_revno is not None:
        try:
            start_rev_id = branch.get_rev_id(start_revno)
        except bzr_errors.NoSuchRevision:
            if not allow_missing:
                raise
            return []
    else:
        start_rev_id = _mod_revision.NULL_REVISION
        start_revno = 0
    if end_revno is None:
        revno, tip_rev_id = branch.last_revision_info()
    else:
        try:
            tip_rev_id = branch.get_rev_id(end_revno)
        except bzr_errors.NoSuchRevision:
            if not allow_missing:
                raise
            return []
        revno = end_revno
    current_revision = branch.repository.get_revision(tip_rev_id)
    revisions = []
    while current_revision is not None:
        if max_revs > -1 and len(revisions) >= max_revs:
            print "Got max revs"
            break
        previous_revision, previous_revno = get_previous_revision(current_revision, revno,
                branch)
        sourcedep_changes = None
        if sourcedeps_path and previous_revision:
            print "Looking in %s for sourcedeps" % sourcedeps_path
            sourcedep_changes = sourcedeps_diff_revisions(branch, current_revision,
                    previous_revision, sourcedeps_path, branch_id,
                    parent_branch_url, fetched_branches=fetched_branches)
            print "Changes in sourcedeps: %s" % sourcedep_changes
        print "Adding rev %s" % revno
        revisions.append((revno, current_revision, sourcedep_changes))
        if (previous_revision is None
            or previous_revision.revision_id == start_rev_id):
            current_revision = None
            revno = None
        else:
            current_revision = previous_revision
            revno = previous_revno
    return revisions


def update_revisions(project_id):
    load_plugins()
    project = Project.objects.get(pk=project_id)
    branch_url = project.branch
    last_revno = project.revision_set.filter(parent__exact=None).aggregate(Max('revno'))['revno__max']
    fetched_branches = {}
    branch = get_branch(branch_url, project.name,
            fetched_branches=fetched_branches)
    try:
        revisions = get_revisions(branch, project.name,
                fixup_branch_location(branch_url.encode('ascii')),
                start_revno=last_revno,
                sourcedeps_path=project.sourcedeps_path,
                fetched_branches=fetched_branches)
    finally:
        branch.unlock()
    store_revisions(project, revisions)


@shared_task
def update_revisions_task(project_id):
    update_revisions(project_id)


@shared_task
def update_all_projects():
    for project in Project.objects.all():
        # Have the subtask expire before the next one will be started to
        # prevent them piling up if the workers pause.
        expires = settings.CELERYBEAT_SCHEDULE['update_all_projects']['schedule'].seconds
        update_revisions_task.subtask((project.pk,), expires=expires).apply_async()
