from datetime import datetime

from django.contrib.auth.models import User
from django.contrib.comments.forms import CommentForm
from django.core.urlresolvers import reverse
from django.http import (
    Http404,
    HttpResponse,
    HttpResponseBadRequest,
    )
from django.middleware.csrf import get_token
from django.shortcuts import (
    get_object_or_404,
    redirect,
    )
from django.views.decorators.http import (
    require_GET,
    require_POST,
    )
import markdown

from . import data
from .api_helpers import (
    ajax_response,
    check_user_allowed,
    )
from .forms import (
    ProjectUpdateRevisionForm,
    RevisionChangeStatusForm,
    SetRollbackForm,
    )
from .models import (
    DeploymentRequest,
    Project,
    Revision,
    )
from .rendering import render_page
from .templating import (
    dump_json_helper,
    get_partials,
    ignore_content,
    include_content,
    render_template,
    )


def project_project_to_info(project):
    info = data.project_basic_info(project)
    info['deployed_revno'] = data.project_deployed_revno(project)
    try:
        deployment_request = project.deployment_request_set.filter(status__in=(DeploymentRequest.STATUS_WAITING,)).get()
        info['deployment_request'] = {
            'url': reverse('deployment_request',
                args=(deployment_request.project.name, deployment_request.pk,)),
            'revno': deployment_request.deploy_revision.revno,
            }
    except DeploymentRequest.DoesNotExist:
        pass
    undeployed_revisions = data.project_undeployed_revisions(project,
        info['deployed_revno'])
    return info, undeployed_revisions


def project_derived_info(undeployed_revisions, child_revisions):
    info = dict()
    deployable_revisions = data.project_deployable_revisions(
            undeployed_revisions, child_revisions)
    info['maximum_deployable_revno'] = data.project_maximum_deployable_revno(
            deployable_revisions)
    revision_blocking_deployment = data.project_revision_blocking_deployment(
            undeployed_revisions, deployable_revisions)
    if revision_blocking_deployment:
        info['revno_blocking_deployment'] = revision_blocking_deployment.revno
    info['undeployed_count'] = len(undeployed_revisions)
    return info


def index_project_to_info(project):
    info = data.project_basic_info(project)
    deployed_revno = data.project_deployed_revno(project)
    undeployed_revisions = data.project_undeployed_revisions(project,
        deployed_revno)
    child_revisions = data.get_child_revisions(undeployed_revisions)
    deployable_revisions = data.project_deployable_revisions(
            undeployed_revisions, child_revisions)
    info['undeployed_count'] = len(undeployed_revisions)
    info['deployable_count'] = len(deployable_revisions)
    info['needing_qa_count'] = len(data.project_revisions_needing_qa(undeployed_revisions))
    return info


def project_revision_to_info(project, revision, child_revisions, comments, show_deployability=None):
    info = data.revision_basic_info(revision)
    summary_status = data.revision_summary_status(revision, child_revisions)
    info['summary_status_display'] = revision.get_summary_status_display(summary_status)
    info['project'] = dict(
        url=reverse('project_view', args=(project.name,)),
        name=project.name,
        )
    if revision.committer:
        info['committer'] = dict(
            url=reverse('person_view', args=(revision.committer.username,)),
            username=revision.committer.username,
            )
    # FIXME: query
    if revision.rolled_back_by:
        info['rolled_back_by_revno'] = revision.rolled_back_by.revno
    info['change_revision_status_url'] = reverse(
            'project_change_revision_status',
            args=(project.name, revision.pk))
    info['set_revision_rollback_url'] = reverse(
            'project_set_revision_rollback',
            args=(project.name, revision.pk))
    info['status_values'] = {
            'Yes': Revision.STATUS_OK,
            'No': Revision.STATUS_BAD,
            'Unknown': Revision.STATUS_UNKNOWN,
            }
    if revision.status == Revision.STATUS_BAD:
        info['no_selected'] = True
        revno = revision.revno
        if revision.parent:
            revno = revision.parent.revno
        # FIXME: query
        info['rollback_candidates'] = list(Revision.objects.filter(revno__gt=revno,
                    project__exact=project,
                    parent__exact=None).values('revno', 'pk'))
    elif revision.status == Revision.STATUS_UNKNOWN:
        info['unknown_selected'] = True
    else:
        info['yes_selected'] = True
    if show_deployability is None:
        info['show_deployability'] = not project.no_qa_of_main_branch
    else:
        info['show_deployability'] = show_deployability
    comment_form = CommentForm(revision)
    info['comment_form'] = comment_form.generate_security_data()
    info['comment_form']['url'] = reverse('comments-post-comment')
    info['comments'] = map(comment_to_info, comments)
    return info


def comment_to_info(comment):
    info = dict()
    info['user'] = comment.userinfo
    info['comment'] = markdown.markdown(comment.comment, safe_mode='escape', extensions=['urlize'])
    info['date'] = comment.submit_date.strftime("%H:%M %Y-%m-%d")
    return info


def project_revision_to_info_with_children(project, revision, child_revisions, comment_map):
    comments = comment_map.get(revision.pk, [])
    info = project_revision_to_info(project, revision, child_revisions, comments)
    def child_info(rev):
        comments = comment_map.get(rev.pk, [])
        return project_revision_to_info(project, rev, [], comments, show_deployability=True)
    info['child_revisions'] = map(child_info, child_revisions)
    return info


def index_data():
    projects = Project.objects.all().select_related('revision_set')
    info = dict()
    info['projects'] = map(index_project_to_info, projects)
    return info


@check_user_allowed
def index(request):
    info = index_data()
    content = render_template('revtracker/index.html', info)
    return HttpResponse(render_page('Deployment status', 'Deployment status',
        content, request.user))


@check_user_allowed
def help(request):
    content = render_template('revtracker/help.html', {})
    return HttpResponse(render_page('Help', 'Help', content, request.user))


def get_project_info(project):
    info = {}
    info['project'], undeployed_revisions = project_project_to_info(project)
    child_revisions = data.get_child_revisions(undeployed_revisions)
    info['project'].update(project_derived_info(undeployed_revisions, child_revisions))
    all_revisions = undeployed_revisions[:]
    for _child_revs in child_revisions.values():
        all_revisions.extend(_child_revs)
    comment_map = data.comments_for_revisions(all_revisions)
    def make_rev_info(rev):
        rev_children = child_revisions.get(rev.pk, [])
        return project_revision_to_info_with_children(project, rev, rev_children, comment_map)
    info['undeployed_revisions'] = map(make_rev_info, undeployed_revisions)
    return info


def get_helpers(request):
   return {
        'csrf_token': csrf_token_helper(request),
        'dump_json': dump_json_helper,
        'current_url': request.path,
        }


def get_project(name):
    try:
        return Project.objects.filter(name__exact=name).select_related('revision_set', 'deployed_revision').get()
    except Project.DoesNotExist:
        raise Http404('No such project: %s' % str(name))


def _update_deployed_revision(request, project):
    success = False
    update_deployed_revision_form = ProjectUpdateRevisionForm(request.POST)
    if update_deployed_revision_form.is_valid():
        try:
            revno = update_deployed_revision_form.cleaned_data['revno']
            deployed_revision = project.revision_set.filter(revno=revno,
                    parent__exact=None).get()
            project.deployed_revision = deployed_revision
            project.save()
            success = True
        except Revision.DoesNotExist:
            update_deployed_revision_form.errors['revno'] = ['Unknown revision']
    errors = update_deployed_revision_form.errors
    return success, errors


def _create_deployment_request(request, project):
    deployment_request = None
    update_deployed_revision_form = ProjectUpdateRevisionForm(request.POST)
    if update_deployed_revision_form.is_valid():
        try:
            revno = update_deployed_revision_form.cleaned_data['revno']
            start_revision = project.deployed_revision
            deploy_revision= project.revision_set.filter(revno=revno,
                    parent__exact=None).get()
            deployment_request = DeploymentRequest.objects.create(
                project=project, start_revision=start_revision,
                deploy_revision=deploy_revision, requester=request.user)
        except Revision.DoesNotExist:
            update_deployed_revision_form.errors['revno'] = ['Unknown revision']
    errors = update_deployed_revision_form.errors
    return deployment_request, errors


@require_POST
@check_user_allowed
def api_v1_project_set_deployed(request, project_name=None):
    project = get_project(project_name)
    errors = dict()
    success, errors = _update_deployed_revision(request, project)
    if success:
        return ajax_response({'status': 'ok'})
    else:
        return ajax_response({'status': 'failed', 'errors': errors})


@require_GET
@check_user_allowed
def api_v1_project(request, project_name=None):
    project = get_project(project_name)
    info = get_project_info(project)
    info['csrf_token'] = get_csrf_token(request)
    return ajax_response(info)


@check_user_allowed
def project(request, project_name=None):
    project = get_project(project_name)
    errors = dict()
    if request.method == "POST":
        deployment_request, errors = _create_deployment_request(request, project)
        if deployment_request is not None:
            return redirect(reverse('deployment_request',
                args=(deployment_request.project.name, deployment_request.pk,)))
    info = get_project_info(project)
    info['project']['errors'] = errors
    helpers = get_helpers(request)
    helpers['if_include_project']= ignore_content
    helpers['if_include_assignee']= include_content
    content = render_template('revtracker/project.html', info,
            get_partials(), helpers)
    return HttpResponse(render_page('Deployment status of %s' % str(project.name),
        project.name, content, request.user, extra_scripts=['project.js']))


def get_deployment_request_info(dr):
    info = dict()
    start_revno = None
    if dr.start_revision:
        start_revno = dr.start_revision.revno
    deploy_revno = dr.deploy_revision.revno
    info['deploy_revno'] = deploy_revno
    info['project'] = {
        'name': dr.project.name,
        'branch': dr.project.branch,
        }
    info['date_requested'] = dr.date_created.strftime("%H:%M %Y-%m-%d")
    info['requester'] = {
        'name': dr.requester.username,
        }
    if dr.date_completed:
        info['date_completed'] = dr.date_completed.strftime("%H:%M %Y-%m-%d")
    if dr.completer:
        info['completer'] = {
            'name': dr.completer.username,
            }
    info['url'] = reverse('deployment_request',
            args=(dr.project.name, dr.pk,))
    deploy_revisions = data.project_revisions_between(dr.project, start_revno, deploy_revno)
    child_revisions = data.get_child_revisions(deploy_revisions)
    all_revisions = deploy_revisions[:]
    for _child_revs in child_revisions.values():
        all_revisions.extend(_child_revs)
    comment_map = data.comments_for_revisions(all_revisions)
    def make_rev_info(rev):
        rev_children = child_revisions.get(rev.pk, [])
        return project_revision_to_info_with_children(dr.project, rev, rev_children, comment_map)
    info['deploy_revisions'] = map(make_rev_info, deploy_revisions)
    def check_deployability(rev):
        rev_children = child_revisions.get(rev.pk, [])
        return data.revision_is_deployable(rev, rev_children)
    bad_revisions = filter(lambda x: not x, map(check_deployability, deploy_revisions))
    if bad_revisions:
        info['not_ok_revisions_count'] = len(bad_revisions)
    info['not_deployable'] = bool(bad_revisions)
    return info


@check_user_allowed
def deployment_request(request, project_name=None, pk=None):
    dr = get_object_or_404(DeploymentRequest,
            project__name__exact=project_name, pk=pk)
    if request.method == "POST":
        dr.status = dr.STATUS_COMPLETE
        dr.date_completed = datetime.utcnow()
        dr.completer = request.user
        dr.project.deployed_revision = dr.deploy_revision
        dr.project.save()
        dr.save()
        return redirect(reverse('project_view', args=(dr.project.name,)))
    info = get_deployment_request_info(dr)
    helpers = get_helpers(request)
    content = render_template('revtracker/deployment_request.html', info,
            get_partials(), helpers)
    return HttpResponse(render_page('Requested deployment of %s' % str(project_name),
        "Deploy revision %d of %s" % (dr.deploy_revision.revno, project_name),
        content, request.user))


def _update_revision_status(revision, request):
    if request.method == "POST":
        form = RevisionChangeStatusForm(request.POST, instance=revision)
        if form.is_valid():
            form.save()
        else:
            return form.errors
    return None


@check_user_allowed
def project_change_revision_status(request, project_name=None, pk=None):
    project = get_object_or_404(Project, name__exact=project_name)
    revision = get_object_or_404(Revision, project=project, pk=pk)
    errors = _update_revision_status(revision, request)
    if errors is None:
        # TODO: tests
        if 'next' in request.GET:
            url = request.GET['next']
        else:
            url = reverse('project_view', args=(revision.project.name,))
        return redirect(url)
    else:
        return HttpResponseBadRequest("Invalid form submission: %s" % errors)


@require_POST
@check_user_allowed
def api_v1_project_change_revision_status(request, project_name=None, pk=None):
    project = get_project(project_name)
    revision = get_object_or_404(Revision, project=project, pk=pk)
    errors = dict()
    errors = _update_revision_status(revision, request)
    if errors is None:
        return ajax_response({'status': 'ok'})
    else:
        return ajax_response({'status': 'failed', 'errors': errors})


def _set_revision_rollback(revision, request):
    if request.method == "POST":
        form = SetRollbackForm(request.POST, instance=revision)
        if form.is_valid():
            form.save()
        else:
            return HttpResponseBadRequest("Invalid form submission: %s" % str(form.errors))
    if request.is_ajax():
        info = get_project_info(revision.project)
        info['csrf_token'] = get_csrf_token(request)
        return ajax_response(info)
    else:
        return redirect(reverse('project_view', args=(revision.project.name,)))


@check_user_allowed
def project_set_revision_rollback(request, project_name=None, pk=None):
    # XXX: tests
    project = get_object_or_404(Project, name__exact=project_name)
    revision = get_object_or_404(Revision, project=project, pk=pk)
    return _set_revision_rollback(revision, request)


def get_csrf_token(request):
    token = get_token(request)
    if token is None:
        # In order to be able to provide debugging info in the
        # case of misconfiguration, we use a sentinel value
        # instead of returning an empty dict.
        return 'NOTPROVIDED'
    else:
        return token


def csrf_token_helper(request):
    def return_csrf_token(this):
        return get_csrf_token(request)
    return return_csrf_token


def get_person_info(person):
    info = dict()
    revisions = Revision.objects.filter(committer__exact=person,
        status__in=(Revision.STATUS_BAD, Revision.STATUS_UNKNOWN)
        ).select_related('committer', 'project')
    child_revisions = data.get_child_revisions(revisions)
    comment_map = data.comments_for_revisions(revisions)
    deployed_revnos = data.deployed_revnos_of_projects_for_revisions(revisions)
    parent_revisions = data.get_parent_revnos(revisions)
    def is_undeployed(rev):
        if rev.project.pk in deployed_revnos:
            parent_revno = parent_revisions.get(rev.pk, rev.revno)
            if parent_revno <= deployed_revnos[rev.project.pk]:
                return False
        return True
    revisions = filter(is_undeployed, revisions)
    def make_rev_info(rev):
        return project_revision_to_info(rev.project, rev,
                child_revisions.get(rev.pk, []), comment_map.get(rev.pk, []),
                show_deployability=True)
    info['revisions_needing_qa'] = map(make_rev_info, revisions)
    return info


@check_user_allowed
def person(request, person_name=None):
    # XXX: tests
    person = get_object_or_404(User, username__exact=person_name)
    # TODO: exclude rolled back revisions that are OK, but include the
    # rollbacks when they are Unknown, i.e. show people rollbacks they
    # should be QAing to fix any BAD revisions they committed
    info = get_person_info(person)
    helpers = get_helpers(request)
    helpers['if_include_project']= include_content
    helpers['if_include_assignee']= ignore_content
    content = render_template('revtracker/person.html', info,
            get_partials(), helpers)
    return HttpResponse(render_page('Revisions committed by %s needing QA' % str(person.username),
        person.username, content, request.user, extra_scripts=['person.js']))


@require_GET
@check_user_allowed
def api_v1_person(request, person_name=None):
    person = get_object_or_404(User, username__exact=person_name)
    info = get_person_info(person)
    info['csrf_token'] = get_csrf_token(request)
    return ajax_response(info)
